/*
  sid (c) 1999-2004 by Dirk Jagdmann <doj@cubic.org>

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  
  $Header: /code/cw/sid/6581.h,v 1.13.2.1 2004/11/05 14:46:40 doj Exp $
*/

#ifndef MOS6581__H
#define MOS6581__H

#include <string.h>

#include "defines.h"
#include "sid.h"

class MOS6581
{
 private:
  MOS6581(const MOS6581&);
  MOS6581& operator= (const MOS6581&);

 protected:
  /// number of SID registers (ok, the real SID doesn't have that much :-)
  static const int MAX=0x20;

  /// proxy of real SID registers
  byte buf[MAX];

  /// sidnumber for unix kernel driver
  int num;

  /// file handle used for SID device
  int fh;

  /// flags which toggle if Voice 1-3 should be on/off
  bool Voice[4], Filter;

  /// backup gate bits used when muting
  bool gateMute[4];

  /// gate trigger detection bits
  bool gateTriggered_[4];

  /// backup filter value used when real Filter is muted
  byte filter;

  /// true if unix device supports IRQ delay/block
  bool irqAvailable_;

  /// true if SID is clocked with PAL frequency, false if NTSC
  bool pal_;

  /// maximum allowed volume
  int maxvolume;
  /// stored real volume as set by C64
  int realvolume;

  /// transpose all frequencies by this value
  double transpose_;

  /// real frequency of voice n
  byte realfreqlo[4], realfreqhi[4];

  /// set frequency in SID, possibly altered by transpose_
  void setfreq(const int v, const byte lo, const byte hi);

  /**
     mute voice i.
     @param i sid voice. 0<i<4
  */
  void mute(int i);

  /**
     unmute voice i.
     @param i sid voice. 0<i<4
  */
  void unmute(int i);

  /// seek filehandle to pos
  void seek(int pos);

  /// write register i into sid device
  void write(const int i);

 public:
  /**
     construct a new sid. It is accessing SID number 0 in the unix device.
  */
  MOS6581();

  virtual ~MOS6581();

  /// set this SID to n
  void setNum(const int n);

  /** uses unix device dev to access real SID chip.
      @param dev filename of unix device. 0 is standard devices should be used.
      @return the return code of the ::open() function
  */
  const int open(const char *dev=0);

  /** toggles state of Voice v.
      @param v sid voice. 0<v<4
      @return the new state of Voice v
  */
  const bool toggle(const int v);

  /** toggles state of Filter
      @return the new state of Filter
  */
  const bool toggleFilter();

  /**
     check if voice is active, thus not muted.
     @param v sid voice or 0 for Filter. 0<=v<4
  */
  bool voiceActive(const int v);

  /// clocks 6581 with ntsc speed
  const bool ntsc();

  /// clocks 6581 with pal speed
  const bool pal();

  /// mutes the 6581
  void mute();

  /// activates all voices and filter again
  void unmute();

  /// @return the file handle of the sidblaster card
  const int sidfh() const { return fh; }

  /**
     set a register
     @param i register of sid, 0<=i<0x19
     @param b value of register
  */
  void set(const int i, byte b);

  /**
     get register value
     @param i register of sid, 0<=i<0x1C
     @return value of register
  */
  byte get(const int i);

  /// reset the SID
  void reset();

  /// set all SID registers to s
  void set(SID& s);

  /// set s to current SID registers
  void get(SID& s);

  /// set volume to v
  void setVolume(const int v=15);

  /// maximum allowed volume to v
  void maxVolume(const int v=15);

  /// set gate detection bits to b
  void setGateBits(const bool b=false);

  /**
     check if ADSR for voice v was started.
     @param v voice number, 0<v<4
     @return true if ADSR was started since last call to setGateBits(false)
  */
  bool gateTriggered(const int v);

  /// @return true if unix device supports IRQ delay/block
  bool irqAvailable() const { return irqAvailable_; }

  /**
     set IRQ delay value to delay us.
     @param delay time in us
     @return 0 on success, -1 on error
  */
  int irqDelay(const int delay);

  /// set transpose to t
  void transpose(const double t) { transpose_=t; }
};

#endif
