/*
  sid (c) 1999-2004 by Dirk Jagdmann <doj@cubic.org>

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  
  $Header: /code/cw/sid/sidinfo.cc,v 1.7 2004/05/14 14:51:04 doj Exp $
*/

extern char *errstr;

#include <curses.h>
#include <stdlib.h>

#include "c64.h"
#include "defines.h"
#include "sid.h"

char* freq2note(const double f)
{
  const double c0=16.35159783;		// frequency (Hz) of the note C0
  static char str[4]={0,0,0,0};
  static char* notes[] = 
    {
      "c-",
      "c#",
      "d-",
      "d#",
      "e-",
      "f-",
      "f#",
      "g-",
      "g#",
      "a-",
      "a#",
      "b-",
    };

  // check for low bound
  if(f<c0)
    {
      str[0]='l';
      str[1]='o';
      str[2]='w';
      return str;
    }

  // get the octave number
  int octave=-1;
  double n1=c0;
  do {
    n1*=2.0;
    octave++;
  } while(f>n1);
  n1*=.5;			// n1 is not base octave

  // check for high bound
  if(octave>9)
    {
      str[0]='h';
      str[1]='i';
      str[2]='!';
      return str;
    }

  // get frequency
  int note=-1;
  n1/=1.029302237; // 2^(1/24)
  do {
    n1*=WURZ12OF2;
    note++;
  } while(f>n1);

  if(note>11)
    note=11;

  // construct return string
  str[0]=notes[note][0];
  str[1]=notes[note][1];
  str[2]=octave+'0';

  return str;
}

static bool initialized=false;
static WINDOW *voicew=0, *sidw=0;

static void cleanup()
{
  if(voicew)
    delwin(voicew), voicew=0;
  if(sidw)
    delwin(sidw), sidw=0;
}

void sidinfo_init(int& xoffset, int& yoffset, C64 *c64)
{
  if(!initialized)
    {
      atexit(cleanup);
      initialized=true;
    }

  // delete any old windows
  cleanup();

  const int sidnum=c64->sidNum();

  // voice window
  voicew=newwin(1+sidnum*3, 40, yoffset, xoffset);
  if(voicew==0)
    {
      errstr="could not create voicew. maybe terminal width too small?";
      exit(1);
    }
  yoffset+=1+sidnum*3;
  wattron(voicew, A_UNDERLINE);
  mvwprintw(voicew, 0, 0, "freq nte pls adsr trsg f waveform");
  wattroff(voicew, A_UNDERLINE);

  // sid window
  sidw=newwin(1+sidnum, 45, yoffset, xoffset);
  if(sidw==0)
    {
      errstr="could not create sidw. maybe terminal width too small?";
      exit(1);
    }
  yoffset+=1+sidnum;
  wattron(sidw, A_UNDERLINE);
  mvwprintw(sidw, 0, 0, "vl 3of freq res ext filter mode   px py o3 e3");
  wattroff(sidw, A_UNDERLINE);

  xoffset+=45;
}

void sidinfo_print(C64 *c64)
{
  allSID allsid;
  c64->get(allsid);
  const bool pal=c64->getPal();
  int wy=1;
  for(int s=0; s<c64->sidNum(); s++)
    {
      SID *sid=&allsid.sid[s];

      // misc window
      wattron(sidw, A_BOLD);
      mvwprintw(sidw, 1+s, 0, "%02i  %c  ", sid->vol, sid->off3?BITtrue:BITfalse);

      if(!c64->voiceActive(0))
	wattroff(sidw, A_BOLD);
      wprintw(sidw, "%04x  %1x   %c  %s %s %s ",
	      ((sid->filtHi)<<3) + (sid->filtLo&7),
	      sid->resonance,
	      sid->filtEx?BITtrue:BITfalse,
	      sid->highpass?"high":"    ",
	      sid->bandpass?"band":"    ",
	      sid->lowpass?"low":"   ");
      if(!c64->voiceActive(0))
	wattron(sidw, A_BOLD);

      wprintw(sidw, "%02x %02x %02x %02x", sid->potx, sid->poty, sid->osc3, sid->env3);
      wattroff(sidw, A_BOLD);
      
      // voices
      for(int i=0; i<3; i++)
	{
	  struct SIDvoice *v=&sid->voice[i];
	  const int chanFreq=lohi(v->freqLo, v->freqHi);
	  const double freq=(double)chanFreq * (pal?985248.0:1.0227E6) / 16777216.0;
	  const int chanPulse=lohi(v->pwLo, v->pwHi&15);
	  bool chanFilt=false;
	  switch(i)
	    {
	    case 0: if(sid->filt1) chanFilt=true; break;
	    case 1: if(sid->filt2) chanFilt=true; break;
	    case 2: if(sid->filt3) chanFilt=true; break;
	    }

	  if(c64->voiceActive(s*3+i+1))
	    wattron(voicew, A_BOLD);
	  mvwprintw(voicew, wy++, 0, "%04x %s %03x %1x%1x%1x%1x %c%c%c%c %c %s %s %s %s",
		    chanFreq,
		    freq2note(freq),
		    chanPulse,
		    v->attack,
		    v->decay,
		    v->sustain,
		    v->release,
		    v->test?BITtrue:BITfalse,
		    v->ring?BITtrue:BITfalse,
		    v->sync?BITtrue:BITfalse,
		    v->gate?BITtrue:BITfalse,
		    chanFilt?BITtrue:BITfalse,
		    v->nse?"nse":"   ",
		    v->squ?"squ":"   ",
		    v->saw?"saw":"   ",
		    v->tri?"tri":"   "
		    );
	  wattroff(voicew, A_BOLD);
	  
	}
    }

  // refresh ncurses
  if(wnoutrefresh(voicew)!=OK)
    {
      errstr="error on refresh of voicew";
      exit(1);
    }
  if(wnoutrefresh(sidw)!=OK)
    {
      errstr="error on refresh of sidw";
      exit(1);
    }
}
